@@ -2,6 +2,10 @@ import argparse
 import re
 import glob
 import os.path
+import signal
+import sys
+from json import JSONDecodeError
+from loguru import logger
 
 import requests
 import semver
@@ -11,14 +15,10 @@ from rich import print
 from rich.console import Console
 
 from .client import (
-    login_to_server,
-    upload,
-    get_server_version_info,
-    get_regex_pattern,
     get_hash_from_checksum_file,
-    check_token,
     get_hash_of_file,
     find_checksum_file,
+    Client,
 )
 from .config import get_config_value, set_config_value, get_optional_true_config_value
 from .constants import (
@@ -26,72 +26,90 @@ from .constants import (
     SERVER_COMPAT_ERROR_MSG,
     SHIPPY_COMPAT_ERROR_MSG,
     SHIPPY_OUTDATED_MSG,
+    UNEXPECTED_SERVER_RESPONSE_ERROR_MSG,
+    FAILED_TO_LOG_IN_ERROR_MSG,
+    CANNOT_CONTACT_SERVER_ERROR_MSG,
+    PRERELEASE_WARNING_MSG,
+    NO_CONFIGURATION_WARNING_MSG,
 )
 from .exceptions import LoginException, UploadException
 from .helper import input_yn, print_error, print_warning, print_success
 from .version import __version__, server_compat_version
 
-ignore_errors = [KeyboardInterrupt, ConnectionError]
-
 sentry_sdk.init(
     SENTRY_SDK_URL,
     traces_sample_rate=1.0,
     release=__version__,
-    ignore_errors=ignore_errors,
+    ignore_errors=[ConnectionError],
 )
 
 console = Console()
 
 
+# Handle SIGINT gracefully and don't puke up traceback
+def sigint_handler(*_):
+    exit(1)
+
+
+signal.signal(signal.SIGINT, sigint_handler)
+
+
+def lower_logger_level():
+    logger.remove()
+    logger.add(sys.stderr, level="INFO")
+
+
 def main():
-    print(f"Welcome to shippy (v.{__version__})!")
-
-    # Check if we cannot prompt the user (default to auto-upload)
-    upload_without_prompt = get_optional_true_config_value(
-        "shippy", "UploadWithoutPrompt"
-    )
-
     # Get commandline arguments
     args = init_argparse()
-    upload_without_prompt = upload_without_prompt or args.yes
+
+    lower_logger_level()
+
+    if args.debug:
+        print_warning("Debug mode has been turned on!")
+        logger.add(sink="shippy_{time}.log", level="DEBUG", enqueue=True)
+
+    print(f"Welcome to shippy (v.{__version__})!")
 
     # Check for updates
     check_shippy_update()
 
-    # Check if server config is valid
-    try:
-        server_url = get_config_value("shippy", "server")
-        if not check_server_url_schema(server_url):
-            print_error(
-                msg=f"The configuration file is corrupt. Please delete it and restart shippy.",
-                newline=True,
-                exit_after=True,
-            )
-        
-        token = get_config_value("shippy", "token")
+    # Initialize client
+    client = build_client_from_config()
+    server_prechecks(client)
 
-        token = check_token_validity(server_url, token)
-    except KeyError:
-        print_warning(
-            "No configuration file found or configuration is invalid. You need to "
-            "configure shippy before you can start using it."
-        )
-        server_url = get_server_url()
-        token = get_token(server_url)
+    # Start uploads
+    search_and_upload_builds(client, args)
 
-        # In case login function updated server URL, we need to fetch it again
-        server_url = get_config_value("shippy", "server")
 
-    # Check server version compatibility
-    check_server_compat(server_url)
+def server_prechecks(client):
+    check_server_compat(client)
+    check_token_validity(client)
 
-    # Get regex pattern from server
-    regex_pattern = get_regex_pattern(server_url=server_url, token=token)
 
-    # Search current directory for files
-    builds = get_builds_in_current_dir(regex_pattern)
+def check_and_upload_build(client, args, build_path):
+    # Check build file validity
+    if not check_build(build_path):
+        print_warning("Invalid build. Skipping...")
+        return
 
-    if len(builds) == 0:
+    if is_upload_without_prompt_enabled(args) or input_yn(
+        f"Uploading build {build_path}. Start?"
+    ):
+        try:
+            upload_id = client.upload(build_path=build_path)
+
+            if is_build_disabling_enabled():
+                client.disable_build(upload_id=upload_id)
+        except UploadException as exception:
+            print_error(exception, newline=True, exit_after=False)
+
+
+def search_and_upload_builds(client, args):
+    # Search current directory for files with regex pattern returned by server
+    build_paths = get_builds_in_current_dir(client.get_regex_pattern())
+
+    if len(build_paths) == 0:
         print_error(
             msg="No files matching the submission criteria were detected in the "
             "current directory.",
@@ -99,26 +117,43 @@ def main():
             exit_after=False,
         )
     else:
-        print(f"Detected {len(builds)} build(s):")
-        for build in builds:
-            print(f"\t{build}")
+        print(f"Detected {len(build_paths)} build(s):")
+        for build_path in build_paths:
+            print(f"\t{build_path}")
 
-        if not upload_without_prompt and len(builds) > 1:
+        if not is_upload_without_prompt_enabled(args) and len(build_paths) > 1:
             print_warning("You seem to be uploading multiple builds. ", newline=False)
             if not input_yn("Are you sure you want to continue?", default=False):
                 return
 
-        for build in builds:
-            # Check build file validity
-            if not check_build(build):
-                print_warning("Invalid build. Skipping...")
-                continue
+        for build_path in build_paths:
+            check_and_upload_build(client, args, build_path)
 
-            if upload_without_prompt or input_yn(f"Uploading build {build}. Start?"):
-                try:
-                    upload(server_url=server_url, build_file_path=build, token=token)
-                except UploadException as exception:
-                    print_error(exception, newline=True, exit_after=False)
+
+def is_upload_without_prompt_enabled(args):
+    config_value = get_optional_true_config_value("shippy", "UploadWithoutPrompt")
+
+    return config_value or args.yes
+
+
+def build_client_from_config():
+    try:
+        url = get_config_value("shippy", "server")
+        if not check_server_url_schema(url):
+            print_error(
+                msg="The configuration file is corrupt. Please delete it and restart "
+                "shippy.",
+                newline=True,
+                exit_after=True,
+            )
+
+        token = get_config_value("shippy", "token")
+        server = Client(server_url=url, token=token)
+    except KeyError:
+        print_warning(NO_CONFIGURATION_WARNING_MSG)
+        server = Client(server_url=get_server_url())
+        prompt_login(server)
+    return server
 
 
 def init_argparse():
@@ -131,50 +166,61 @@ def init_argparse():
         action="store_true",
         help="Upload builds automatically without prompting",
     )
+    parser.add_argument(
+        "-d",
+        "--debug",
+        action="store_true",
+        help="Enable debug mode",
+    )
     return parser.parse_args()
 
 
-def check_server_compat(server_url):
+def check_server_compat(client):
     with console.status(
         "Please wait while shippy contacts the remote server to check compatibility... "
     ):
-        server_version_info = get_server_version_info(server_url)
+        # Check if shipper version is compatible
+        if not client.is_server_compatible():
+            print_error(
+                msg=SERVER_COMPAT_ERROR_MSG.format(
+                    client.get_version(), server_compat_version
+                ),
+                newline=True,
+                exit_after=True,
+            )
 
-    # Check if shipper version is compatible
-    if semver.compare(server_version_info["version"], server_compat_version) == -1:
-        print_error(
-            msg=SERVER_COMPAT_ERROR_MSG.format(
-                server_version_info["version"], server_compat_version
-            ),
-            newline=True,
-            exit_after=True,
-        )
-
-    # Check if shippy version is compatible
-    if semver.compare(server_version_info["shippy_compat_version"], __version__) == 1:
-        print_error(
-            msg=SHIPPY_COMPAT_ERROR_MSG.format(
-                server_version_info["shippy_compat_version"], __version__
-            ),
-            newline=True,
-            exit_after=True,
+    # Check if shippy version is compatible, but only if running stable builds
+    if is_prerelease():
+        print_warning(
+            "You're running a prerelease build of shippy. Server compatibility checks "
+            "are disabled."
         )
+    else:
+        if not client.is_shippy_compatible():
+            print_error(
+                msg=SHIPPY_COMPAT_ERROR_MSG.format(
+                    client.get_shippy_compat_version(), __version__
+                ),
+                newline=True,
+                exit_after=True,
+            )
 
     print_success("Finished compatibility check. No problems found.")
 
 
-def check_token_validity(server_url, token):
+def check_token_validity(client):
     with console.status(
         "Please wait while shippy contacts the remote server to check if the token is "
         "still valid... "
     ):
-        is_token_valid = check_token(server_url, token)
-
-    if not is_token_valid:
-        # Token check failed, prompt for login again
-        print_warning("The saved token is invalid. Please sign-in again.")
-        token = get_token(server_url)
-    return token
+        if client.is_token_valid():
+            print_success(
+                f"Successfully validated token! Hello, {client.get_username()}!"
+            )
+        else:
+            # Token check failed, prompt for login again
+            print_warning("The saved token is invalid. Please sign-in again.")
+            prompt_login(client)
 
 
 def check_shippy_update():
@@ -184,10 +230,19 @@ def check_shippy_update():
         )
         latest_version = r.json()["name"]
 
-    if semver.compare(__version__, latest_version) == -1:
-        print(SHIPPY_OUTDATED_MSG.format(__version__, latest_version))
+    # Check if user is running an alpha/beta build
+    if is_prerelease():
+        print_warning(PRERELEASE_WARNING_MSG)
     else:
-        print_success("Finished update check. shippy is up-to-date!")
+        # User is running a stable build, proceed with update check
+        if semver.compare(__version__, latest_version) == -1:
+            print(SHIPPY_OUTDATED_MSG.format(__version__, latest_version))
+        else:
+            print_success("Finished update check. shippy is up-to-date!")
+
+
+def is_prerelease():
+    return "a" in __version__ or "b" in __version__
 
 
 def get_builds_in_current_dir(regex_pattern):
@@ -204,7 +259,6 @@ def get_builds_in_current_dir(regex_pattern):
 def check_build(filename):
     """Makes sure the build is valid"""
     print(f"Validating build {filename}...")
-
     # Validate that there is a matching checksum file
     has_checksum_file_type, has_sum_postfix = find_checksum_file(filename=filename)
 
@@ -215,10 +269,7 @@ def check_build(filename):
         return False
 
     # Validate checksum
-    with console.status(
-        f"Checking {has_checksum_file_type.upper()} hash of {filename}... this may "
-        "take a couple of seconds. "
-    ):
+    with console.status(f"Checking {has_checksum_file_type.upper()} hash..."):
         hash_val = get_hash_of_file(
             filename=filename, checksum_type=has_checksum_file_type
         )
@@ -237,7 +288,7 @@ def check_build(filename):
                 exit_after=False,
             )
             return False
-        print_success(f"{has_checksum_file_type.upper()} hash of {filename} matched.")
+        print_success("Success!")
 
     build_slug, _ = os.path.splitext(filename)
     _, _, _, build_type, build_variant, _ = build_slug.split("-")
@@ -251,7 +302,9 @@ def check_build(filename):
     valid_variants = ["gapps", "vanilla", "foss", "goapps"]
     if build_variant not in valid_variants:
         print_error(
-            msg="This build has an unknown variant. ", newline=False, exit_after=False
+            msg="This build has an unknown variant. ",
+            newline=False,
+            exit_after=False,
         )
         return False
 
@@ -287,7 +340,7 @@ def check_server_url_schema(url):
     return "http" in url
 
 
-def get_token(server_url):
+def prompt_login(client):
     while True:
         from getpass import getpass
 
@@ -296,16 +349,33 @@ def get_token(server_url):
             password = getpass(prompt="Enter your password: ")
 
             try:
-                token = login_to_server(username, password, server_url)
-                set_config_value("shippy", "token", token)
-                return token
+                client.login(username=username, password=password)
+                set_config_value("shippy", "token", client.token)
             except LoginException as exception:
+                print_error(exception, newline=True, exit_after=True)
+            except JSONDecodeError:
                 print_error(
-                    f"{exception} Please try again.", newline=True, exit_after=False
+                    msg=UNEXPECTED_SERVER_RESPONSE_ERROR_MSG
+                    + FAILED_TO_LOG_IN_ERROR_MSG,
+                    newline=True,
+                    exit_after=True,
+                )
+            except requests.exceptions.RequestException:
+                print_error(
+                    msg=CANNOT_CONTACT_SERVER_ERROR_MSG + FAILED_TO_LOG_IN_ERROR_MSG,
+                    newline=True,
+                    exit_after=True,
                 )
         except KeyboardInterrupt:
             exit(0)
 
 
+def is_build_disabling_enabled():
+    try:
+        return get_config_value("shippy", "DisableBuildOnUpload") == "true"
+    except KeyError:
+        return False
+
+
 if __name__ == "__main__":
     main()
